==> azorius-0.3.2.tgz: azorius-0.3.2/files.go <==
//
// Copyright (c) 2023 Ted Unangst <tedu@tedunangst.com>
//
// Permission to use, copy, modify, and distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

package main

import (
	"crypto/sha512"
	"database/sql"
	"fmt"
	"io/fs"
	"os"
	"path"
)

type fileInfo struct {
	xid   string
	media string
}

func hashfiledata(data []byte) string {
	h := sha512.New512_256()
	h.Write(data)
	return fmt.Sprintf("%x", h.Sum(nil))
}

func attachmentfromurl(url string, longterm bool) (*Attachment, error) {
	info, ok := findfileurl(url)
	if ok {
		att := Attachment{
			Xid:   info.xid,
			Media: info.media,
		}
		return &att, nil
	}
	data, err := fetchData(url)
	if err != nil {
		return nil, err
	}
	return attachmentfromdata(data, url, longterm)
}

func attachmentfromdata(data []byte, url string, longterm bool) (*Attachment, error) {
	img, err := shrinkit(data)
	if err != nil {
		return nil, err
	}
	data = img.Data
	format := img.Format
	media := "image/" + format
	if format == "jpeg" {
		format = "jpg"
	}
	if format == "svg+xml" {
		format = "svg"
	}
	xid, err := savefiledata(data, format, media, url, longterm)
	if err != nil {
		return nil, err
	}
	att := Attachment{
		Xid:   xid,
		Media: media,
	}
	return &att, nil
}

func attachmentfromxid(xid string) (*Attachment, error) {
	info, ok := findfile(xid)
	if !ok {
		return nil, fmt.Errorf("lost file for xid")
	}
	att := Attachment{
		Xid:   info.xid,
		Media: info.media,
	}
	return &att, nil
}

func fileexists(filename string) bool {
	_, err := os.Stat(filename)
	return err == nil
}

func longtermname(xid string) string {
	return fmt.Sprintf("%s/attachments/%s/%s", dataDir, xid[:3], xid)
}

func shorttermname(xid string) string {
	return fmt.Sprintf("%s/tempstore/%s/%s", dataDir, xid[:3], xid)
}

func savefiledata(data []byte, format, media, url string, longterm bool) (string, error) {
	h := hashfiledata(data)
	infos := findfilehashes(h)
	for _, info := range infos {
		if fileexists(longtermname(info.xid)) {
			return info.xid, nil
		}
		if !longterm && fileexists(shorttermname(info.xid)) {
			return info.xid, nil
		}
	}
	xid := generateDailyXid() + "." + format
	dirname := dataDir
	if longterm {
		dirname += "/attachments/"
	} else {
		dirname += "/tempstore/"
	}
	dirname += xid[:3]
	os.Mkdir(dirname, 0755)
	filename := dirname + "/" + xid
	fd, err := os.Create(filename)
	if err != nil {
		return "", err
	}
	defer fd.Close()
	fd.Write(data)
	err = dbSaveFileInfo(h, url, media, xid)
	if err != nil {
		elog.Printf("error saving file info: %s", err)
	}
	return xid, nil
}

func scanfileinfo(row RowLike) (fileInfo, bool) {
	var info fileInfo
	err := row.Scan(&info.xid, &info.media)
	if err != nil {
		if err != sql.ErrNoRows {
			elog.Printf("error scanning file xid: %s", err)
		}
		return info, false
	}
	return info, true
}

func findfilehashes(h string) []fileInfo {
	var infos []fileInfo
	rows, err := stmtFindFileHash.Query(h)
	if err != nil {
		elog.Printf("error finding file hashes: %s", err)
		return infos
	}
	defer rows.Close()
	for rows.Next() {
		info, ok := scanfileinfo(rows)
		if !ok {
			continue
		}
		infos = append(infos, info)
	}
	return infos
}

func findfileurl(url string) (fileInfo, bool) {
	row := stmtFindFileUrl.QueryRow(url)
	return scanfileinfo(row)
}

func findfile(xid string) (fileInfo, bool) {
	row := stmtFindFile.QueryRow(xid)
	return scanfileinfo(row)
}

func dbSaveFileInfo(h, url, media, xid string) error {
	_, err := stmtSaveFileInfo.Exec(h, url, media, xid)
	return err
}

func previewfile(fullname string) string {
	ext := path.Ext(fullname)
	prevname := fullname[:len(fullname)-len(ext)] + "-small" + ext
	if fileexists(prevname) {
		return prevname
	}
	dlog.Printf("generating preview: %s", prevname)
	data, err := os.ReadFile(fullname)
	if err != nil {
		return fullname
	}
	img, err := previewit(data)
	if err != nil {
		return fullname
	}
	fd, err := os.Create(prevname)
	if err != nil {
		return fullname
	}
	defer fd.Close()
	fd.Write(img.Data)
	fd.Close()
	return prevname
}

func markFiles(rows *sql.Rows, err error, files map[string]bool, used bool) error {
	if err != nil {
		return err
	}
	defer rows.Close()
	for rows.Next() {
		var xid string
		err = rows.Scan(&xid)
		if err != nil {
			elog.Printf("error scanning file xid: %s", err)
			continue
		}
		files[xid] = used
	}
	return nil
}

func filescleanup() {
	db := opendatabase()
	tx, err := db.Begin()
	if err != nil {
		elog.Printf("error starting transaction")
	}
	files := make(map[string]bool)
	rows, err := tx.Query("select xid from files")
	err = markFiles(rows, err, files, false)
	if err != nil {
		elog.Printf("error selecting files: %s", err)
		return
	}
	rows, err = tx.Query("select xid from attachments")
	err = markFiles(rows, err, files, true)
	if err != nil {
		elog.Printf("error selecting attachments: %s", err)
		return
	}
	for xid, used := range files {
		if !used {
			_, err = tx.Exec("delete from files where xid = ?", xid)
			if err != nil {
				elog.Printf("error deleting unused file: %s", err)
				continue
			}
		}
	}
	err = tx.Commit()
	if err != nil {
		elog.Printf("error committing tx: %s", err)
	}

	var dead []string
	walker := func(pathname string, ent fs.DirEntry, err error) error {
		if ent.IsDir() {
			return nil
		}
		if !files[path.Base(pathname)] {
			dead = append(dead, pathname)
		}
		return nil
	}
	dir := os.DirFS(dataDir)
	fs.WalkDir(dir, "attachments", walker)
	fs.WalkDir(dir, "tempstore", walker)
	trash := dataDir + "/trash"
	os.Mkdir(trash, 0755)
	trash += "/"
	for _, file := range dead {
		err := os.Rename(file, trash+path.Base(file))
		if err != nil {
			elog.Printf("error moving %s to trash: %s", file, err)
		}
	}
}

==> azorius-0.4.0.tgz: azorius-0.4.0/files.go <==
//
// Copyright (c) 2023 Ted Unangst <tedu@tedunangst.com>
//
// Permission to use, copy, modify, and distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

package main

import (
	"crypto/sha512"
	"database/sql"
	"fmt"
	"io/fs"
	"os"
	"path"
	"strings"
	"sync"
)

type fileInfo struct {
	xid   string
	media string
}

func hashfiledata(data []byte) string {
	h := sha512.New512_256()
	h.Write(data)
	return fmt.Sprintf("%x", h.Sum(nil))
}

func attachmentfromurl(url string, longterm bool) (*Attachment, error) {
	info, ok := findfileurl(url)
	if ok {
		att := Attachment{
			Xid:   info.xid,
			Media: info.media,
		}
		return &att, nil
	}
	data, err := fetchData(url)
	if err != nil {
		return nil, err
	}
	return attachmentfromdata(data, url, longterm)
}

func attachmentfromdata(data []byte, url string, longterm bool) (*Attachment, error) {
	img, err := shrinkit(data)
	if err != nil {
		return nil, err
	}
	data = img.Data
	format := img.Format
	media := "image/" + format
	if format == "jpeg" {
		format = "jpg"
	}
	if format == "svg+xml" {
		format = "svg"
	}
	xid, err := savefiledata(data, format, media, url, longterm)
	if err != nil {
		return nil, err
	}
	att := Attachment{
		Xid:   xid,
		Media: media,
	}
	return &att, nil
}

func attachmentfromxid(xid string) (*Attachment, error) {
	info, ok := findfile(xid)
	if !ok {
		return nil, fmt.Errorf("lost file for xid")
	}
	att := Attachment{
		Xid:   info.xid,
		Media: info.media,
	}
	return &att, nil
}

func fileexists(filename string) bool {
	_, err := os.Stat(filename)
	return err == nil
}

func longtermname(xid string) string {
	return fmt.Sprintf("%s/attachments/%s/%s", dataDir, xid[:3], xid)
}

func shorttermname(xid string) string {
	return fmt.Sprintf("%s/tempstore/%s/%s", dataDir, xid[:3], xid)
}

func savefiledata(data []byte, format, media, url string, longterm bool) (string, error) {
	h := hashfiledata(data)
	infos := findfilehashes(h)
	for _, info := range infos {
		if fileexists(longtermname(info.xid)) {
			return info.xid, nil
		}
		if !longterm && fileexists(shorttermname(info.xid)) {
			return info.xid, nil
		}
	}
	xid := generateDailyXid() + "." + format
	dirname := dataDir
	if longterm {
		dirname += "/attachments/"
	} else {
		dirname += "/tempstore/"
	}
	dirname += xid[:3]
	os.Mkdir(dirname, 0755)
	filename := dirname + "/" + xid
	fd, err := os.Create(filename)
	if err != nil {
		return "", err
	}
	defer fd.Close()
	fd.Write(data)
	err = dbSaveFileInfo(h, url, media, xid)
	if err != nil {
		elog.Printf("error saving file info: %s", err)
	}
	return xid, nil
}

func scanfileinfo(row RowLike) (fileInfo, bool) {
	var info fileInfo
	err := row.Scan(&info.xid, &info.media)
	if err != nil {
		if err != sql.ErrNoRows {
			elog.Printf("error scanning file xid: %s", err)
		}
		return info, false
	}
	return info, true
}

func findfilehashes(h string) []fileInfo {
	var infos []fileInfo
	rows, err := stmtFindFileHash.Query(h)
	if err != nil {
		elog.Printf("error finding file hashes: %s", err)
		return infos
	}
	defer rows.Close()
	for rows.Next() {
		info, ok := scanfileinfo(rows)
		if !ok {
			continue
		}
		infos = append(infos, info)
	}
	return infos
}

func findfileurl(url string) (fileInfo, bool) {
	row := stmtFindFileUrl.QueryRow(url)
	return scanfileinfo(row)
}

func findfile(xid string) (fileInfo, bool) {
	row := stmtFindFile.QueryRow(xid)
	return scanfileinfo(row)
}

func dbSaveFileInfo(h, url, media, xid string) error {
	_, err := stmtSaveFileInfo.Exec(h, url, media, xid)
	return err
}

func previewfile(fullname string) string {
	ext := path.Ext(fullname)
	prevname := fullname[:len(fullname)-len(ext)] + "-small" + ext
	if fileexists(prevname) {
		return prevname
	}
	dlog.Printf("generating preview: %s", prevname)
	data, err := os.ReadFile(fullname)
	if err != nil {
		return fullname
	}
	img, err := previewit(data)
	if err != nil {
		return fullname
	}
	fd, err := os.Create(prevname)
	if err != nil {
		return fullname
	}
	defer fd.Close()
	fd.Write(img.Data)
	fd.Close()
	return prevname
}

func markFiles(rows *sql.Rows, err error, files map[string]bool, used bool) error {
	if err != nil {
		return err
	}
	xids := scanstrings(make([]string, 0, 256), rows, err)
	for _, xid := range xids {
		files[xid] = used
	}
	return nil
}

var filesToClean []string
var mtxToClean sync.Mutex

func cleansomefiles(xids []string) {
	if len(xids) == 0 {
		return
	}
	mtxToClean.Lock()
	defer mtxToClean.Unlock()
	filesToClean = append(filesToClean, xids...)
	if len(filesToClean) > 100 {
		cleansomefilesnow(filesToClean)
		filesToClean = nil
	}
}

func cleansomefilesnow(xids []string) {
	db := opendatabase()
	tx, err := db.Begin()
	if err != nil {
		elog.Printf("error starting cleanfiles transaction")
		return
	}
	defer tx.Rollback()
	files := make(map[string]bool)
	for _, xid := range xids {
		files[xid] = false
	}
	q := fmt.Sprintf("select xid from attachments where xid in ('%s')", strings.Join(xids, "','"))
	rows, err := tx.Query(q)
	err = markFiles(rows, err, files, true)
	if err != nil {
		elog.Printf("error selecting attachments: %s", err)
		return
	}
	for xid, used := range files {
		if !used {
			_, err = tx.Exec("delete from files where xid = ?", xid)
			if err != nil {
				elog.Printf("error deleting unused file: %s", err)
			}
			dlog.Printf("deleting unused file %s", xid)
			os.Remove(longtermname(xid))
			os.Remove(shorttermname(xid))
		}
	}
	err = tx.Commit()
	if err != nil {
		elog.Printf("error committing cleanfiles tx: %s", err)
	}
}

func cleanallfiles() {
	db := opendatabase()
	tx, err := db.Begin()
	if err != nil {
		elog.Printf("error starting cleanfiles transaction")
		return
	}
	defer tx.Rollback()
	files := make(map[string]bool)
	rows, err := tx.Query("select xid from files")
	err = markFiles(rows, err, files, false)
	if err != nil {
		elog.Printf("error selecting files: %s", err)
		return
	}
	rows, err = tx.Query("select xid from attachments")
	err = markFiles(rows, err, files, true)
	if err != nil {
		elog.Printf("error selecting attachments: %s", err)
		return
	}
	for xid, used := range files {
		if !used {
			_, err = tx.Exec("delete from files where xid = ?", xid)
			if err != nil {
				elog.Printf("error deleting unused file: %s", err)
				continue
			}
		}
	}
	err = tx.Commit()
	if err != nil {
		elog.Printf("error committing cleanfiles tx: %s", err)
	}

	var dead []string
	walker := func(pathname string, ent fs.DirEntry, err error) error {
		if ent.IsDir() {
			return nil
		}
		if !files[path.Base(pathname)] {
			dead = append(dead, pathname)
		}
		return nil
	}
	dir := os.DirFS(dataDir)
	fs.WalkDir(dir, "attachments", walker)
	fs.WalkDir(dir, "tempstore", walker)
	trash := dataDir + "/trash"
	os.Mkdir(trash, 0755)
	trash += "/"
	for _, file := range dead {
		err := os.Rename(file, trash+path.Base(file))
		if err != nil {
			elog.Printf("error moving %s to trash: %s", file, err)
		}
	}
}
